@@ -2,7 +2,7 @@
 
 Copyright (c) 1996, 2017, Oracle and/or its affiliates. All rights reserved.
 Copyright (c) 2009, Percona Inc.
-Copyright (c) 2013, 2022, MariaDB Corporation.
+Copyright (c) 2013, 2023, MariaDB Corporation.
 
 Portions of this file contain modifications contributed and copyrighted
 by Percona Inc.. Those modifications are
@@ -62,10 +62,8 @@ Created 2/16/1996 Heikki Tuuri
 #include "btr0btr.h"
 #include "btr0cur.h"
 #include "rem0rec.h"
-#include "ibuf0ibuf.h"
 #include "srv0start.h"
 #include "srv0srv.h"
-#include "btr0defragment.h"
 #include "mysql/service_wsrep.h" /* wsrep_recovery */
 #include "trx0rseg.h"
 #include "buf0flu.h"
@@ -88,6 +86,7 @@ Created 2/16/1996 Heikki Tuuri
 #include "row0row.h"
 #include "row0mysql.h"
 #include "btr0pcur.h"
+#include "ibuf0ibuf.h"
 #include "zlib.h"
 #include "log.h"
 
@@ -198,7 +197,7 @@ static dberr_t create_log_file(bool create_new_db, lsn_t lsn)
 	os_file_t file{
           os_file_create_func(logfile0.c_str(),
                               OS_FILE_CREATE,
-                              OS_LOG_FILE, false, &ret)
+                              OS_FILE_NORMAL, OS_LOG_FILE, false, &ret)
         };
 
 	if (!ret) {
@@ -222,12 +221,7 @@ close_and_exit:
 	if (!log_sys.attach(file, srv_log_file_size)) {
 		goto close_and_exit;
 	}
-
-	mysql_mutex_lock(&recv_sys.mutex);
-	const bool all_opened = fil_system.sys_space->open(create_new_db);
-	mysql_mutex_unlock(&recv_sys.mutex);
-
-	if (!all_opened) {
+	if (!fil_system.sys_space->open(create_new_db)) {
 		goto err_exit;
 	}
 
@@ -243,16 +237,7 @@ close_and_exit:
 		srv_startup_is_before_trx_rollback_phase = false;
 	}
 
-	/* Enable log_checkpoint() in buf_flush_page_cleaner().
-	If we are upgrading or resizing the log at startup, we must not
-	write any FILE_MODIFY or FILE_CHECKPOINT records to the old log file
-	because the ib_logfile0 could be in an old format that we can only
-	check for emptiness, or the write could lead to an overflow.
-
-	At this point, we are not holding recv_sys.mutex, but there are no
-	pending page reads, and buf_pool.flush_list is empty. Therefore,
-	we can clear the flag without risking any race condition with
-	buf_page_t::read_complete(). */
+	/* Enable checkpoints in buf_flush_page_cleaner(). */
 	recv_sys.recovery_on = false;
 	log_sys.latch.wr_unlock();
 
@@ -296,7 +281,7 @@ static dberr_t srv_undo_tablespace_create(const char* name)
 		innodb_data_file_key,
 		name,
 		srv_read_only_mode ? OS_FILE_OPEN : OS_FILE_CREATE,
-		OS_DATA_FILE, srv_read_only_mode, &ret);
+		OS_FILE_NORMAL, OS_DATA_FILE, srv_read_only_mode, &ret);
 
 	if (!ret) {
 		if (os_file_get_last_error(false) != OS_FILE_ALREADY_EXISTS
@@ -471,7 +456,8 @@ ATTRIBUTE_COLD static dberr_t srv_undo_tablespaces_reinit()
       buf_block_t *block= buf_page_get_gen(
         rseg->page_id(), 0, RW_X_LATCH, nullptr, BUF_GET, &mtr);
       if (!block) break;
-      while (!fseg_free_step(block, TRX_RSEG + TRX_RSEG_FSEG_HEADER, &mtr));
+      while (!fseg_free_step(TRX_RSEG + TRX_RSEG_FSEG_HEADER +
+                             block->page.frame, &mtr));
     }
   }
 
@@ -610,7 +596,8 @@ static uint32_t trx_rseg_get_n_undo_tablespaces()
   mtr_t mtr;
   mtr.start();
 
-  if (const buf_block_t *sys_header= trx_sysf_get(&mtr, false))
+  if (const buf_block_t *sys_header=
+      recv_sys.recover({TRX_SYS_SPACE, TRX_SYS_PAGE_NO}, &mtr, nullptr))
     for (ulint rseg_id= 0; rseg_id < TRX_SYS_N_RSEGS; rseg_id++)
       if (trx_sysf_rseg_get_page_no(sys_header, rseg_id) != FIL_NULL)
 	if (uint32_t space= trx_sysf_rseg_get_space(sys_header, rseg_id))
@@ -648,7 +635,7 @@ static uint32_t srv_undo_tablespace_open(bool create, const char* name,
 
   pfs_os_file_t fh= os_file_create(innodb_data_file_key, name,
                                    OS_FILE_OPEN_SILENT,
-                                   OS_DATA_FILE,
+                                   OS_FILE_AIO, OS_DATA_FILE,
                                    srv_read_only_mode, &success);
 
   if (!success)
@@ -657,14 +644,13 @@ static uint32_t srv_undo_tablespace_open(bool create, const char* name,
   ulint n_retries = 5;
   os_offset_t size= os_file_get_size(fh);
   ut_a(size != os_offset_t(-1));
-  page_t *apage= nullptr;
-  const page_t *page= nullptr;
 
   if (!create)
   {
-    apage= static_cast<byte*>(aligned_malloc(srv_page_size, srv_page_size));
+    page_t *page= static_cast<byte*>(aligned_malloc(srv_page_size,
+                                                    srv_page_size));
 undo_retry:
-    if (os_file_read(IORequestRead, fh, apage, 0, srv_page_size, nullptr) !=
+    if (os_file_read(IORequestRead, fh, page, 0, srv_page_size, nullptr) !=
         DB_SUCCESS)
     {
 err_exit:
@@ -675,12 +661,11 @@ err_exit:
         n_retries--;
         goto undo_retry;
       }
-      sql_print_error("InnoDB: Unable to read first page of file %s", name);
-      aligned_free(apage);
+      ib::error() << "Unable to read first page of file " << name;
+      aligned_free(page);
       return ~0U;
     }
 
-    page= apage;
     DBUG_EXECUTE_IF("undo_space_read_fail", goto err_exit;);
 
     uint32_t id= mach_read_from_4(FIL_PAGE_SPACE_ID + page);
@@ -688,30 +673,18 @@ err_exit:
         memcmp_aligned<2>(FIL_PAGE_SPACE_ID + page,
                           FSP_HEADER_OFFSET + FSP_SPACE_ID + page, 4))
     {
-      sql_print_error("InnoDB: Inconsistent tablespace ID in file %s", name);
+      ib::error() << "Inconsistent tablespace ID in file " << name;
       goto err_exit;
     }
 
     space_id= id;
     fsp_flags= mach_read_from_4(FSP_HEADER_OFFSET + FSP_SPACE_FLAGS + page);
 
-    if (buf_page_is_corrupted(false, page, fsp_flags))
-    {
-      page= recv_sys.dblwr.find_page(page_id_t{space_id, 0}, LSN_MAX);
-      if (!page)
-      {
-        /* If the first page of the given user tablespace is not there
-        in the doublewrite buffer, then the recovery is going to fail
-        now. Report error only when doublewrite buffer is not empty */
-        sql_print_error("InnoDB: Corrupted page "
-                        "[page id: space=" UINT32PF ", page number=0]"
-                        " of datafile '%s' could not be found"
-                        " in the doublewrite buffer", space_id, name);
-        goto err_exit;
-      }
+    if (buf_page_is_corrupted(false, page, fsp_flags) &&
+        recv_sys.dblwr.restore_first_page(space_id, name, fh))
+      goto err_exit;
 
-      fsp_flags= mach_read_from_4(FSP_HEADER_OFFSET + FSP_SPACE_FLAGS + page);
-    }
+    aligned_free(page);
   }
 
   /* Load the tablespace into InnoDB's internal data structures. */
@@ -723,7 +696,8 @@ err_exit:
   fil_set_max_space_id_if_bigger(space_id);
 
   mysql_mutex_lock(&fil_system.mutex);
-  fil_space_t *space= fil_space_t::create(space_id, fsp_flags, false, nullptr,
+  fil_space_t *space= fil_space_t::create(space_id, fsp_flags,
+                                          FIL_TYPE_TABLESPACE, nullptr,
                                           FIL_ENCRYPTION_DEFAULT, true);
   ut_ad(space);
   fil_node_t *file= space->add(name, fh, 0, false, true);
@@ -733,7 +707,7 @@ err_exit:
     space->set_sizes(SRV_UNDO_TABLESPACE_SIZE_IN_PAGES);
     space->size= file->size= uint32_t(size >> srv_page_size_shift);
   }
-  else if (!file->read_page0(page, true))
+  else if (!file->read_page0())
   {
     os_file_close(file->handle);
     file->handle= OS_FILE_CLOSED;
@@ -742,7 +716,6 @@ err_exit:
   }
 
   mysql_mutex_unlock(&fil_system.mutex);
-  aligned_free(apage);
   return space_id;
 }
 
@@ -766,6 +739,7 @@ srv_check_undo_redo_logs_exists()
 		fh = os_file_create_func(
 			name,
 			OS_FILE_OPEN_RETRY_SILENT,
+			OS_FILE_NORMAL,
 			OS_DATA_FILE,
 			srv_read_only_mode,
 			&ret);
@@ -787,7 +761,7 @@ srv_check_undo_redo_logs_exists()
 
 	fh = os_file_create_func(logfilename.c_str(),
 				 OS_FILE_OPEN_RETRY_SILENT,
-				 OS_LOG_FILE,
+				 OS_FILE_NORMAL, OS_LOG_FILE,
 				 srv_read_only_mode, &ret);
 
 	if (ret) {
@@ -900,12 +874,11 @@ dberr_t srv_undo_tablespaces_init(bool create_new_undo, mtr_t *mtr)
   srv_undo_tablespaces_active= srv_undo_tablespaces;
 
   uint32_t n_undo= (create_new_undo || srv_operation == SRV_OPERATION_BACKUP ||
-                    srv_operation == SRV_OPERATION_RESTORE_DELTA)
+                 srv_operation == SRV_OPERATION_RESTORE_DELTA)
     ? srv_undo_tablespaces : TRX_SYS_N_RSEGS;
 
-  mysql_mutex_lock(&recv_sys.mutex);
-  dberr_t err= srv_all_undo_tablespaces_open(create_new_undo, n_undo);
-  mysql_mutex_unlock(&recv_sys.mutex);
+  if (dberr_t err= srv_all_undo_tablespaces_open(create_new_undo, n_undo))
+    return err;
 
   /* Initialize srv_undo_space_id_start=0 when there are no
   dedicated undo tablespaces. */
@@ -913,11 +886,16 @@ dberr_t srv_undo_tablespaces_init(bool create_new_undo, mtr_t *mtr)
     srv_undo_space_id_start= 0;
 
   if (create_new_undo)
-    for (uint32_t i= 0; err == DB_SUCCESS && i < srv_undo_tablespaces; ++i)
-      err= fsp_header_init(fil_space_get(srv_undo_space_id_start + i),
-                           SRV_UNDO_TABLESPACE_SIZE_IN_PAGES, mtr);
+  {
+    for (uint32_t i= 0; i < srv_undo_tablespaces; ++i)
+    {
+      dberr_t err= fsp_header_init(fil_space_get(srv_undo_space_id_start + i),
+                                   SRV_UNDO_TABLESPACE_SIZE_IN_PAGES, mtr);
+      if (err) return err;
+    }
+  }
 
-  return err;
+  return DB_SUCCESS;
 }
 
 /** Create the temporary file tablespace.
@@ -975,19 +953,12 @@ srv_open_tmp_tablespace(bool create_new_db)
 	return(err);
 }
 
-/** Shutdown background threads, except the page cleaner.
-@param init_abort set to true when InnoDB startup aborted */
-static void srv_shutdown_threads(bool init_abort= false)
+/** Shutdown background threads, except the page cleaner. */
+static void srv_shutdown_threads()
 {
 	ut_ad(!srv_undo_sources);
 	srv_master_timer.reset();
-	/* In case of InnoDB start up aborted, Don't change
-	the srv_shutdown_state. Because innodb_shutdown()
-	does call innodb_preshutdown() which changes the
-	srv_shutdown_state back to SRV_SHUTDOWN_INITIATED */
-	if (!init_abort) {
-		srv_shutdown_state = SRV_SHUTDOWN_EXIT_THREADS;
-	}
+	srv_shutdown_state = SRV_SHUTDOWN_EXIT_THREADS;
 
 	if (purge_sys.enabled()) {
 		srv_purge_shutdown();
@@ -1057,14 +1028,14 @@ srv_init_abort_low(
 	}
 
 	srv_shutdown_bg_undo_sources();
-	srv_shutdown_threads(true);
+	srv_shutdown_threads();
 	return(err);
 }
 
 /** Prepare to delete the redo log file. Flush the dirty pages from all the
 buffer pools.  Flush the redo log buffer to the redo log file.
 @return lsn upto which data pages have been flushed. */
-static lsn_t srv_prepare_to_delete_redo_log_file() noexcept
+ATTRIBUTE_COLD static lsn_t srv_prepare_to_delete_redo_log_file()
 {
   DBUG_ENTER("srv_prepare_to_delete_redo_log_file");
 
@@ -1078,7 +1049,7 @@ static lsn_t srv_prepare_to_delete_redo_log_file() noexcept
 
   log_sys.latch.wr_lock(SRW_LOCK_CALL);
   const bool latest_format{log_sys.is_latest()};
-  lsn_t flushed_lsn{log_sys.get_flushed_lsn(std::memory_order_relaxed)};
+  lsn_t flushed_lsn{log_sys.get_lsn()};
 
   if (latest_format && !(log_sys.file_size & 4095) &&
       flushed_lsn != log_sys.next_checkpoint_lsn +
@@ -1086,11 +1057,6 @@ static lsn_t srv_prepare_to_delete_redo_log_file() noexcept
        ? SIZE_OF_FILE_CHECKPOINT + 8
        : SIZE_OF_FILE_CHECKPOINT))
   {
-#ifdef HAVE_PMEM
-    if (!log_sys.is_opened())
-      log_sys.buf_size= unsigned(std::min<uint64_t>(log_sys.capacity(),
-                                                    log_sys.buf_size_max));
-#endif
     fil_names_clear(flushed_lsn);
     flushed_lsn= log_sys.get_lsn();
   }
@@ -1128,10 +1094,9 @@ same_size:
 
   log_sys.latch.wr_unlock();
 
-  if (latest_format)
-    log_write_up_to(flushed_lsn, false);
+  log_write_up_to(flushed_lsn, false);
 
-  ut_ad(flushed_lsn == log_get_lsn());
+  ut_ad(flushed_lsn == log_sys.get_lsn());
   ut_ad(!os_aio_pending_reads());
   ut_d(mysql_mutex_lock(&buf_pool.flush_list_mutex));
   ut_ad(!buf_pool.get_oldest_modification(0));
@@ -1141,23 +1106,82 @@ same_size:
   DBUG_RETURN(flushed_lsn);
 }
 
+/** Upgrade the redo log to the latest format, or change its size
+or encryption, before starting to write any log records. */
+ATTRIBUTE_COLD static dberr_t srv_log_rebuild()
+{
+  /* Prepare to delete the old redo log file */
+  const lsn_t lsn{srv_prepare_to_delete_redo_log_file()};
+
+  DBUG_EXECUTE_IF("innodb_log_abort_1", return DB_ERROR;);
+  /* Prohibit redo log writes from any other threads until creating a
+  log checkpoint at the end of create_log_file(). */
+  ut_d(recv_no_log_write= true);
+  ut_ad(!os_aio_pending_reads());
+  ut_d(mysql_mutex_lock(&buf_pool.flush_list_mutex));
+  ut_ad(!buf_pool.get_oldest_modification(0));
+  ut_d(mysql_mutex_unlock(&buf_pool.flush_list_mutex));
+  /* os_aio_pending_writes() may hold here if some write_io_callback()
+  did not release the slot yet.  However, the page write itself must
+  have completed, because the buf_pool.flush_list is empty. In debug
+  builds, we wait for this to happen, hoping to get a hung process if
+  this assumption does not hold. */
+  ut_d(os_aio_wait_until_no_pending_writes(false));
+
+  /* Close the redo log file, so that we can replace it */
+  log_sys.close_file();
+
+  DBUG_EXECUTE_IF("innodb_log_abort_5", return DB_ERROR;);
+
+  dberr_t err= create_log_file(false, lsn);
+
+  if (err == DB_SUCCESS && log_sys.resize_rename())
+    err = DB_ERROR;
+
+  return err;
+}
+
+/** Rebuild the redo log if needed. */
+static dberr_t srv_log_rebuild_if_needed()
+{
+  if (srv_force_recovery == SRV_FORCE_NO_LOG_REDO)
+    /* Completely ignore the redo log. */
+    return DB_SUCCESS;
+  if (srv_read_only_mode)
+    /* Leave the redo log alone. */
+    return DB_SUCCESS;
+
+  if (log_sys.file_size == srv_log_file_size &&
+      log_sys.format ==
+      (srv_encrypt_log ? log_t::FORMAT_ENC_10_8 : log_t::FORMAT_10_8))
+  {
+    /* No need to add or remove encryption, upgrade, or resize. */
+    delete_log_files();
+    return DB_SUCCESS;
+  }
+
+  return srv_log_rebuild();
+}
+
+ATTRIBUTE_COLD static dberr_t ibuf_log_rebuild_if_needed()
+{
+  mysql_mutex_lock(&recv_sys.mutex);
+  recv_sys.apply(true);
+  mysql_mutex_unlock(&recv_sys.mutex);
+
+  if (recv_sys.is_corrupt_log() || recv_sys.is_corrupt_fs())
+    return DB_CORRUPTION;
+
+  dberr_t err= srv_log_rebuild_if_needed();
+  recv_sys.debug_free();
+  return err;
+}
+
 static tpool::task_group rollback_all_recovered_group(1);
 static tpool::task rollback_all_recovered_task(trx_rollback_all_recovered,
 					       nullptr,
 					       &rollback_all_recovered_group);
 
-inline lsn_t log_t::init_lsn() noexcept
-{
-  latch.wr_lock(SRW_LOCK_CALL);
-  ut_ad(!write_lsn_offset);
-  write_lsn_offset= 0;
-  const lsn_t lsn{base_lsn.load(std::memory_order_relaxed)};
-  flushed_to_disk_lsn.store(lsn, std::memory_order_relaxed);
-  write_lsn= lsn;
-  latch.wr_unlock();
-  return lsn;
-}
-
 /** Start InnoDB.
 @param[in]	create_new_db	whether to create a new database
 @return DB_SUCCESS or error code */
@@ -1180,7 +1204,7 @@ dberr_t srv_start(bool create_new_db)
 
 	if (srv_read_only_mode) {
 		sql_print_information("InnoDB: Started in read only mode");
-		srv_use_doublewrite_buf = false;
+		buf_dblwr.use = buf_dblwr.USE_NO;
 	}
 
 	high_level_read_only = srv_read_only_mode
@@ -1195,10 +1219,6 @@ dberr_t srv_start(bool create_new_db)
 	ib::info() << "!!!!!!!! UNIV_DEBUG switched on !!!!!!!!!";
 #endif
 
-#ifdef UNIV_IBUF_DEBUG
-	ib::info() << "!!!!!!!! UNIV_IBUF_DEBUG switched on !!!!!!!!!";
-#endif
-
 	ib::info() << "Compressed tables use zlib " ZLIB_VERSION
 #ifdef UNIV_ZIP_DEBUG
 	      " with validation"
@@ -1288,24 +1308,56 @@ dberr_t srv_start(bool create_new_db)
 		return(srv_init_abort(err));
 	}
 
-	if (srv_read_only_mode) {
-		ib::info() << "Disabling background log and ibuf IO write"
-			<< " threads.";
-	}
-
 	if (os_aio_init()) {
+		ib::error() << "Cannot initialize AIO sub-system";
+
 		return(srv_init_abort(DB_ERROR));
 	}
 
+#ifdef LINUX_NATIVE_AIO
+	if (srv_use_native_aio) {
+		ib::info() << "Using Linux native AIO";
+	}
+#endif
+#ifdef HAVE_URING
+	if (srv_use_native_aio) {
+		ib::info() << "Using liburing";
+	}
+#endif
+
 	fil_system.create(srv_file_per_table ? 50000 : 5000);
 
+	ib::info() << "Initializing buffer pool, total size = "
+		<< ib::bytes_iec{srv_buf_pool_size}
+		<< ", chunk size = " << ib::bytes_iec{srv_buf_pool_chunk_unit};
+
 	if (buf_pool.create()) {
+		ib::error() << "Cannot allocate memory for the buffer pool";
+
 		return(srv_init_abort(DB_ERROR));
 	}
 
-	log_sys.create();
+	ib::info() << "Completed initialization of buffer pool";
+
+#ifdef UNIV_DEBUG
+	/* We have observed deadlocks with a 5MB buffer pool but
+	the actual lower limit could very well be a little higher. */
+
+	if (srv_buf_pool_size <= 5 * 1024 * 1024) {
+
+		ib::info() << "Small buffer pool size ("
+			<< ib::bytes_iec{srv_buf_pool_size}
+			<< "), the flst_validate() debug function can cause a"
+			<< " deadlock if the buffer pool fills up.";
+	}
+#endif /* UNIV_DEBUG */
+
+	if (!log_sys.create()) {
+		return srv_init_abort(DB_ERROR);
+	}
+
 	recv_sys.create();
-	lock_sys.create(srv_lock_table_size = 5 * buf_pool.curr_size());
+	lock_sys.create(srv_lock_table_size);
 
 	srv_startup_is_before_trx_rollback_phase = true;
 
@@ -1378,14 +1430,9 @@ dberr_t srv_start(bool create_new_db)
 
 	/* Open data files in the system tablespace: we keep
 	them open until database shutdown */
-	mysql_mutex_lock(&recv_sys.mutex);
 	ut_d(fil_system.sys_space->recv_size = srv_sys_space_size_debug);
-	err = fil_system.sys_space->open(create_new_db)
-		? DB_SUCCESS : DB_ERROR;
-	mysql_mutex_unlock(&recv_sys.mutex);
 
-
-	if (err == DB_SUCCESS) {
+	if (fil_system.sys_space->open(create_new_db)) {
 		mtr_t mtr;
 		mtr.start();
 		err= srv_undo_tablespaces_init(create_new_db, &mtr);
@@ -1415,31 +1462,41 @@ dberr_t srv_start(bool create_new_db)
 	if (create_new_db) {
 		ut_ad(!srv_read_only_mode);
 
-		mtr_start(&mtr);
+		mtr.start();
 		ut_ad(fil_system.sys_space->id == 0);
 		compile_time_assert(TRX_SYS_SPACE == 0);
-		compile_time_assert(IBUF_SPACE_ID == 0);
-		ut_a(fsp_header_init(fil_system.sys_space,
-				     uint32_t(sum_of_new_sizes), &mtr)
-		     == DB_SUCCESS);
-
-		ulint ibuf_root = btr_create(
-			DICT_CLUSTERED | DICT_IBUF, fil_system.sys_space,
-			DICT_IBUF_ID_MIN, nullptr, &mtr, &err);
-
-		mtr_commit(&mtr);
-
-		if (ibuf_root == FIL_NULL) {
-			return srv_init_abort(err);
+		err = fsp_header_init(fil_system.sys_space,
+				      uint32_t(sum_of_new_sizes), &mtr);
+		/* Allocate dummy change buffer pages for backward
+		compatibility and to prevent a downgrade. */
+		if (err != DB_SUCCESS) {
+		} else if (buf_block_t *b =
+			   fseg_create(fil_system.sys_space, PAGE_DATA, &mtr,
+				       &err)) {
+			ut_ad(b->page.id()
+			      == page_id_t(0, FSP_IBUF_HEADER_PAGE_NO));
+			b = fseg_alloc_free_page_general(
+				b->page.frame + PAGE_DATA,
+				FSP_IBUF_TREE_ROOT_PAGE_NO, FSP_UP, false,
+				&mtr, &mtr, &err);
+			if (b) {
+				ut_ad(b->page.id() == page_id_t
+				      (0, FSP_IBUF_TREE_ROOT_PAGE_NO));
+				mtr.set_modified(*b);
+				fsp_init_file_page(fil_system.sys_space, b,
+						   &mtr);
+			} else {
+				ut_ad(err != DB_SUCCESS);
+			}
 		}
-
-		ut_ad(ibuf_root == IBUF_TREE_ROOT_PAGE_NO);
-
 		/* To maintain backward compatibility we create only
 		the first rollback segment before the double write buffer.
 		All the remaining rollback segments will be created later,
 		after the double write buffer has been created. */
-		err = trx_sys_create_sys_pages(&mtr);
+		if (err == DB_SUCCESS) {
+			err = trx_sys_create_sys_pages(&mtr);
+		}
+		mtr.commit();
 
 		if (err != DB_SUCCESS) {
 			return(srv_init_abort(err));
@@ -1471,39 +1528,47 @@ dberr_t srv_start(bool create_new_db)
 		err = recv_recovery_from_checkpoint_start();
 		recv_sys.close_files();
 
-		if (err != DB_SUCCESS) {
-			return(srv_init_abort(err));
-		}
+		recv_sys.dblwr.pages.clear();
+
+		bool must_upgrade_ibuf = false;
 
 		switch (srv_operation) {
                 case SRV_OPERATION_NORMAL:
 		case SRV_OPERATION_EXPORT_RESTORED:
 		case SRV_OPERATION_RESTORE_EXPORT:
-			/* Initialize the change buffer. */
-			err = dict_boot();
 			if (err != DB_SUCCESS) {
-				return(srv_init_abort(err));
+				break;
 			}
+
+			err = ibuf_upgrade_needed();
+
+			if (UNIV_UNLIKELY(err == DB_FAIL)) {
+				must_upgrade_ibuf = true;
+				err = ibuf_log_rebuild_if_needed();
+			}
+
+			if (err != DB_SUCCESS) {
+				break;
+			}
+
+			err = dict_boot();
 			/* fall through */
 		case SRV_OPERATION_RESTORE:
-			/* This must precede recv_sys.apply(true). */
+			if (err != DB_SUCCESS) {
+				break;
+			}
+
 			srv_undo_tablespaces_active
 				= trx_rseg_get_n_undo_tablespaces();
-
-			if (srv_operation != SRV_OPERATION_RESTORE) {
-				dict_sys.load_sys_tables();
-			}
-			err = trx_lists_init_at_db_start();
-			if (err != DB_SUCCESS) {
-				return srv_init_abort(err);
-			}
 			break;
-		case SRV_OPERATION_RESTORE_DELTA:
-		case SRV_OPERATION_BACKUP:
-		case SRV_OPERATION_BACKUP_NO_DEFER:
+		default:
 			ut_ad("wrong mariabackup mode" == 0);
 		}
 
+		if (err != DB_SUCCESS) {
+			return srv_init_abort(err);
+		}
+
 		if (srv_force_recovery < SRV_FORCE_NO_LOG_REDO) {
 			/* Apply the hashed log records to the
 			respective file pages, for the last batch of
@@ -1525,9 +1590,45 @@ dberr_t srv_start(bool create_new_db)
 				return(srv_init_abort(DB_CORRUPTION));
 			}
 
+			if (srv_operation != SRV_OPERATION_RESTORE
+			    || recv_needed_recovery) {
+			}
+
 			DBUG_PRINT("ib_log", ("apply completed"));
 
-			if (recv_needed_recovery) {
+			if (srv_operation != SRV_OPERATION_RESTORE) {
+				dict_sys.lock(SRW_LOCK_CALL);
+				dict_load_sys_table(dict_sys.sys_tables);
+				dict_sys.unlock();
+
+				if (UNIV_UNLIKELY(must_upgrade_ibuf)) {
+					dict_load_tablespaces(nullptr, true);
+					err = ibuf_upgrade();
+					if (err != DB_SUCCESS) {
+						return srv_init_abort(err);
+					}
+				}
+
+				dict_sys.lock(SRW_LOCK_CALL);
+				dict_load_sys_table(dict_sys.sys_columns);
+				dict_load_sys_table(dict_sys.sys_indexes);
+				dict_load_sys_table(dict_sys.sys_fields);
+				dict_sys.unlock();
+				dict_sys.load_sys_tables();
+
+				err = trx_lists_init_at_db_start();
+				if (err != DB_SUCCESS) {
+					return srv_init_abort(err);
+				}
+
+				if (recv_needed_recovery) {
+					trx_sys_print_mysql_binlog_offset();
+				}
+			} else if (recv_needed_recovery) {
+				err = trx_lists_init_at_db_start();
+				if (err != DB_SUCCESS) {
+					return srv_init_abort(err);
+				}
 				trx_sys_print_mysql_binlog_offset();
 			}
 		}
@@ -1544,57 +1645,10 @@ dberr_t srv_start(bool create_new_db)
 		generating any dirty pages, so that the old redo log
 		file will not be written to. */
 
-		if (srv_force_recovery == SRV_FORCE_NO_LOG_REDO) {
-			/* Completely ignore the redo log. */
-		} else if (srv_read_only_mode) {
-			/* Leave the redo log alone. */
-		} else if (log_sys.file_size == srv_log_file_size
-			   && log_sys.format
-			   == (srv_encrypt_log
-			       ? log_t::FORMAT_ENC_10_8
-			       : log_t::FORMAT_10_8)) {
-			/* No need to add or remove encryption,
-			upgrade, or resize. */
-			delete_log_files();
-		} else {
-			/* Prepare to delete the old redo log file */
-			const lsn_t lsn{srv_prepare_to_delete_redo_log_file()};
+		err = srv_log_rebuild_if_needed();
 
-			DBUG_EXECUTE_IF("innodb_log_abort_1",
-					return(srv_init_abort(DB_ERROR)););
-			/* Prohibit redo log writes from any other
-			threads until creating a log checkpoint at the
-			end of create_log_file(). */
-			ut_d(recv_no_log_write = true);
-			ut_ad(!os_aio_pending_reads());
-			ut_d(mysql_mutex_lock(&buf_pool.flush_list_mutex));
-			ut_ad(!buf_pool.get_oldest_modification(0));
-			ut_d(mysql_mutex_unlock(&buf_pool.flush_list_mutex));
-			/* os_aio_pending_writes() may hold here if
-			some write_io_callback() did not release the
-			slot yet. However, the page write itself must
-			have completed, because the buf_pool.flush_list
-			is empty. In debug builds, we wait for this to
-			happen, hoping to get a hung process if this
-			assumption does not hold. */
-			ut_d(os_aio_wait_until_no_pending_writes(false));
-
-			/* Close the redo log file, so that we can replace it */
-			log_sys.close_file();
-
-			DBUG_EXECUTE_IF("innodb_log_abort_5",
-					return(srv_init_abort(DB_ERROR)););
-			DBUG_PRINT("ib_log", ("After innodb_log_abort_5"));
-
-			err = create_log_file(false, lsn);
-
-			if (err == DB_SUCCESS && log_sys.resize_rename()) {
-				err = DB_ERROR;
-			}
-
-			if (err != DB_SUCCESS) {
-				return(srv_init_abort(err));
-			}
+		if (err != DB_SUCCESS) {
+			return srv_init_abort(err);
 		}
 
 		recv_sys.debug_free();
@@ -1733,8 +1787,7 @@ dberr_t srv_start(bool create_new_db)
 			/* Bitmap page types will be reset in
 			buf_dblwr_check_block() without redo logging. */
 			block = buf_page_get(
-				page_id_t(IBUF_SPACE_ID,
-					  FSP_IBUF_HEADER_PAGE_NO),
+				page_id_t(0, FSP_IBUF_HEADER_PAGE_NO),
 				0, RW_X_LATCH, &mtr);
 			if (UNIV_UNLIKELY(!block)) {
 			corrupted_old_page:
@@ -1792,13 +1845,7 @@ dberr_t srv_start(bool create_new_db)
 		}
 
 		if (srv_force_recovery < SRV_FORCE_NO_UNDO_LOG_SCAN) {
-			/* The following call is necessary for the change
-			buffer to work with multiple tablespaces. We must
-			know the mapping between space id's and .ibd file
-			names.
-
-			We also determine the maximum tablespace id used. */
-			dict_check_tablespaces_and_store_max_id(nullptr);
+			dict_load_tablespaces();
 		}
 
 		if (srv_force_recovery < SRV_FORCE_NO_TRX_UNDO
@@ -1869,21 +1916,18 @@ skip_monitors:
 	if (srv_print_verbose_log) {
 		sql_print_information("InnoDB: "
 				      "log sequence number " LSN_PF
+#ifdef HAVE_PMEM
 				      "%s"
+#endif
 				      "; transaction id " TRX_ID_FMT,
 				      recv_sys.lsn,
-				      log_sys.is_mmap()
+#ifdef HAVE_PMEM
+				      log_sys.is_pmem()
 				      ? " (memory-mapped)" : "",
+#endif
 				      trx_sys.get_max_trx_id());
 	}
 
-	if (srv_force_recovery == 0) {
-		/* In the change buffer we may have even bigger tablespace
-		id's, because we may have dropped those tablespaces, but
-		the buffered records have not been cleaned yet. */
-		ibuf_update_max_tablespace_id();
-	}
-
 	if (!srv_read_only_mode) {
 		if (create_new_db) {
 			srv_buffer_pool_load_at_startup = FALSE;
@@ -1914,9 +1958,6 @@ skip_monitors:
 		fil_crypt_threads_cond. */
 		fil_crypt_threads_init();
 
-		/* Initialize online defragmentation. */
-		btr_defragment_init();
-
 		srv_started_redo = true;
 	}
 
@@ -1937,15 +1978,9 @@ void innodb_preshutdown()
   if (srv_read_only_mode)
     return;
   if (!srv_fast_shutdown && srv_operation <= SRV_OPERATION_EXPORT_RESTORED)
-  {
-    /* Because a slow shutdown must empty the change buffer, we had
-    better prevent any further changes from being buffered. */
-    innodb_change_buffering= 0;
-
     if (srv_force_recovery < SRV_FORCE_NO_TRX_UNDO && srv_was_started)
       while (trx_sys.any_active_transactions())
         std::this_thread::sleep_for(std::chrono::milliseconds(1));
-  }
   srv_shutdown_bg_undo_sources();
   srv_purge_shutdown();
 
@@ -2008,8 +2043,6 @@ void innodb_shutdown()
 	      || srv_force_recovery >= SRV_FORCE_NO_TRX_UNDO);
 	ut_ad(lock_sys.is_initialised() || !srv_was_started);
 	ut_ad(log_sys.is_initialised() || !srv_was_started);
-	ut_ad(ibuf.index || !innodb_change_buffering || !srv_was_started
-	      || srv_force_recovery >= SRV_FORCE_NO_DDL_UNDO);
 
 	dict_stats_deinit();
 
@@ -2019,7 +2052,6 @@ void innodb_shutdown()
 		fts_optimize_shutdown(); dict_stats_shutdown(); */
 
 		fil_crypt_threads_cleanup();
-		btr_defragment_shutdown();
 	}
 
 	/* This must be disabled before closing the buffer pool
@@ -2030,7 +2062,6 @@ void innodb_shutdown()
 		btr_search_disable();
 	}
 #endif /* BTR_CUR_HASH_ADAPT */
-	ibuf_close();
 	log_sys.close();
 	purge_sys.close();
 	trx_sys.close();
